"""
Copied from https://github.com/bbangert/beaker/blob/1.13.0/beaker/util.py

Copyright (c) 2006, 2007 Ben Bangert, Mike Bayer, Philip Jenvey
                         and contributors.
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions
are met:
1. Redistributions of source code must retain the above copyright
   notice, this list of conditions and the following disclaimer.
2. Redistributions in binary form must reproduce the above copyright
   notice, this list of conditions and the following disclaimer in the
   documentation and/or other materials provided with the distribution.
3. The name of the author or contributors may not be used to endorse or
   promote products derived from this software without specific prior
   written permission.

THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
SUCH DAMAGE.
"""

from datetime import datetime, timedelta
import re


def convert(behaviors):
    rules = [
        ('cas', (bool, int), 'cas must be a boolean or an integer'),
        ('no_block', (bool, int), 'no_block must be a boolean or an integer'),
        ('receive_timeout', (int,), 'receive_timeout must be an integer'),
        ('send_timeout', (int,), 'send_timeout must be an integer'),
        (
            'ketama_hash',
            (str,),
            'ketama_hash must be a string designating a valid hashing strategy option',
        ),
        ('_poll_timeout', (int,), '_poll_timeout must be an integer'),
        ('auto_eject', (bool, int), 'auto_eject must be an integer'),
        ('retry_timeout', (int,), 'retry_timeout must be an integer'),
        ('_sort_hosts', (bool, int), '_sort_hosts must be an integer'),
        ('_io_msg_watermark', (int,), '_io_msg_watermark must be an integer'),
        ('ketama', (bool, int), 'ketama must be a boolean or an integer'),
        ('ketama_weighted', (bool, int), 'ketama_weighted must be a boolean or an integer'),
        ('_io_key_prefetch', (int, bool), '_io_key_prefetch must be a boolean or an integer'),
        (
            '_hash_with_prefix_key',
            (bool, int),
            '_hash_with_prefix_key must be a boolean or an integer',
        ),
        ('tcp_nodelay', (bool, int), 'tcp_nodelay must be a boolean or an integer'),
        ('failure_limit', (int,), 'failure_limit must be an integer'),
        ('buffer_requests', (bool, int), 'buffer_requests must be a boolean or an integer'),
        ('_socket_send_size', (int,), '_socket_send_size must be an integer'),
        ('num_replicas', (int,), 'num_replicas must be an integer'),
        ('remove_failed', (int,), 'remove_failed must be an integer'),
        ('_noreply', (bool, int), '_noreply must be a boolean or an integer'),
        ('_io_bytes_watermark', (int,), '_io_bytes_watermark must be an integer'),
        ('_socket_recv_size', (int,), '_socket_recv_size must be an integer'),
        (
            'distribution',
            (str,),
            'distribution must be a string designating a valid distribution option',
        ),
        ('connect_timeout', (int,), 'connect_timeout must be an integer'),
        ('hash', (str,), 'hash must be a string designating a valid hashing option'),
        ('verify_keys', (bool, int), 'verify_keys must be a boolean or an integer'),
        ('dead_timeout', (int,), 'dead_timeout must be an integer'),
    ]

    return verify_rules(behaviors, rules)


def verify_rules(params, ruleset):
    for key, types, message in ruleset:
        if key in params:
            params[key] = verify_options(params[key], types, message)
    return params


def verify_options(opt, types, error):
    if not isinstance(opt, types):
        if not isinstance(types, tuple):
            types = (types,)
        coerced = False
        for typ in types:
            try:
                if typ in (list, tuple):
                    opt = [x.strip() for x in opt.split(',')]
                else:
                    if typ is bool:
                        typ = asbool
                    elif typ is int:
                        typ = asint
                    elif typ in (timedelta, datetime):
                        if not isinstance(opt, typ):
                            raise Exception('%s requires a timedelta type', typ)
                    opt = typ(opt)
                coerced = True
            except Exception:
                pass
            if coerced:
                break
        if not coerced:
            raise Exception(error)
    elif isinstance(opt, str) and not opt.strip():
        raise Exception('Empty strings are invalid for: %s' % error)
    return opt


def asbool(obj):
    if isinstance(obj, str):
        obj = obj.strip().lower()
        if obj in ['true', 'yes', 'on', 'y', 't', '1']:
            return True
        elif obj in ['false', 'no', 'off', 'n', 'f', '0']:
            return False
        else:
            raise ValueError('String is not true/false: %r' % obj)
    return bool(obj)


def asint(obj):
    if isinstance(obj, int):
        return obj
    elif isinstance(obj, str) and re.match(r'^\d+$', obj):
        return int(obj)
    else:
        raise Exception('This is not a proper int')



from dogpile.cache.backends.memory import MemoryBackend
import dogpile.cache.backends.memcached
import dogpile.cache.exception
import pytest

from pyramid_dogpile_cache2 import configure_dogpile_cache, get_region


def test_sets_backend_for_regions(empty_config):
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo, bar',
            'dogpile_cache.backend': 'dogpile.cache.memory',
            'dogpile_cache.expiration_time': '1',
        }
    )
    assert isinstance(get_region('foo').backend, MemoryBackend)
    assert isinstance(get_region('bar').backend, MemoryBackend)


def test_converts_expiration_time(empty_config):
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.memory',
            'dogpile_cache.expiration_time': '1',
        }
    )
    assert get_region('foo').expiration_time == 1


def test_converts_expiration_time_to_int(empty_config):
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.memory',
            'dogpile_cache.expiration_time': '1',
        }
    )
    assert get_region('foo').expiration_time == 1


def test_sets_memcache_expire_time_to_later_than_expiration_time(empty_config, monkeypatch):
    monkeypatch.setattr(
        dogpile.cache.backends.memcached.PylibmcBackend, '_imports', lambda self: None
    )
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.pylibmc',
            'dogpile_cache.pylibmc_url': 'http://localhost:8899',
            'dogpile_cache.expiration_time': '1',
            'dogpile_cache.memcache_expire_time_interval': '5',
        }
    )
    assert get_region('foo').backend.set_arguments['time'] == 6


def test_sets_pylibmc_behaviours(empty_config, monkeypatch):
    monkeypatch.setattr(
        dogpile.cache.backends.memcached.PylibmcBackend, '_imports', lambda self: None
    )
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.pylibmc',
            'dogpile_cache.pylibmc_url': 'http://localhost:8899',
            'dogpile_cache.expiration_time': '1',
            'dogpile_cache.pylibmc_behavior.send_timeout': '5',
        }
    )
    assert get_region('foo').backend.behaviors == {'send_timeout': 5}


def test_unconfigured_region_raises(empty_config):
    get_region('bar')
    with pytest.raises(dogpile.cache.exception.RegionNotConfigured):
        configure_dogpile_cache(
            {
                'dogpile_cache.regions': 'foo',
                'dogpile_cache.backend': 'dogpile.cache.memory',
                'dogpile_cache.expiration_time': '1',
            }
        )


def test_multiple_calls_clear_backend(empty_config):
    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.memory',
            'dogpile_cache.expiration_time': '1',
        }
    )

    region = get_region('foo')

    # Init cached value in region
    _ = region.actual_backend

    original_backend = region.backend
    assert region.actual_backend == original_backend

    configure_dogpile_cache(
        {
            'dogpile_cache.regions': 'foo',
            'dogpile_cache.backend': 'dogpile.cache.memory',
            'dogpile_cache.expiration_time': '1',
        }
    )

    # Confirm region is back to fresh state
    new_backend = region.backend
    assert new_backend != original_backend
    assert region._actual_backend is None

    # Regen cached value. Confirm cache points to new backend
    _ = region.actual_backend
    assert region.actual_backend == new_backend



# coding: utf-8
import dogpile.cache.util

from pyramid_dogpile_cache2 import get_region
import pyramid_dogpile_cache2.cache


def test_key_generator_handles_type_annotations():
    def function_with_annotations() -> None:
        pass

    gen = pyramid_dogpile_cache2.cache.key_generator(None, function_with_annotations)
    assert gen() == 'pyramid_dogpile_cache2.tests.test_cache.function_with_annotations|'


def test_mangle_key_handles_non_ascii_arguments():
    pyramid_dogpile_cache2.cache.sha1_mangle_key('föö')


def test_key_generator_adds_class_name_for_methods():
    def plain_function():
        pass

    class Foo(object):
        def method(self):
            pass

    gen = dogpile.cache.util.function_key_generator(None, plain_function)
    assert gen() == 'pyramid_dogpile_cache2.tests.test_cache:plain_function|'
    gen = dogpile.cache.util.function_key_generator(None, Foo().method)
    assert gen() == 'pyramid_dogpile_cache2.tests.test_cache:method|'

    gen = pyramid_dogpile_cache2.cache.key_generator(None, plain_function)
    assert gen() == 'pyramid_dogpile_cache2.tests.test_cache.plain_function|'
    foo = Foo()
    gen = pyramid_dogpile_cache2.cache.key_generator(None, foo.method)
    assert gen(foo) == 'pyramid_dogpile_cache2.tests.test_cache.Foo.method|'


def test_methods_are_detected_when_decorated(empty_config):
    region = get_region('test')
    region.configure('dogpile.cache.memory')

    class Foo(object):
        @region.cache_on_arguments()
        def method(self):
            return 42

    Foo().method()
    assert region.get('pyramid_dogpile_cache2.tests.test_cache.Foo.method|') == 42



import dogpile.cache
import dogpile.cache.backends.memory
import dogpile.cache.exception


CACHE_REGIONS = {}


def get_region(name):
    """This is the main entry point to the caching system; it returns a
    CacheRegion that can be used to cache function results.

    Example usage::

        CACHE = pyramid_dogpile_cache2.get_region('longterm')
        @CACHE.cache_on_arguments()
        def expensive_function(one, two, three):
            # compute stuff
    """
    # Prevent import cycle
    from pyramid_dogpile_cache2.cache import key_generator, sha1_mangle_key

    if name not in CACHE_REGIONS:
        CACHE_REGIONS[name] = dogpile.cache.make_region(
            name, function_key_generator=key_generator, key_mangler=sha1_mangle_key
        )
    return CACHE_REGIONS[name]


def clear():
    """Removes all cached values from all configured in-memory cache regions.
    Useful for unit tests.
    """
    for region in CACHE_REGIONS.values():
        backend = getattr(region, 'backend', None)
        if isinstance(backend, dogpile.cache.backends.memory.MemoryBackend):
            region.backend._cache.clear()


def includeme(config):
    configure_dogpile_cache(config.registry.settings)


def configure_dogpile_cache(settings):
    # This is somewhat inspired by pyramid_dogpile_cache.includeme()
    # and build_dogpile_region_from_dict().

    # Prevent import cycle
    from pyramid_dogpile_cache2.settings import build_dogpile_region_settings

    _parse_dogpile_cache_settings(settings)
    _, region_settings = build_dogpile_region_settings(settings)

    for name, settings in region_settings.items():
        # XXX Type conversion for all sorts of settings is woefully incomplete.
        if 'expiration_time' in settings:
            settings['expiration_time'] = exp = int(settings['expiration_time'])
            settings.setdefault(
                'arguments.memcached_expire_time',
                exp + int(settings.get('memcache_expire_time_interval', 30)),
            )

        region = get_region(name)
        # XXX kludgy: Remove any existing backend configuration, so
        # configure_dogpile_cache() may be called multiple times (which
        # should only happen in tests).
        region.__dict__.pop('backend', None)
        # Pop cached actual_backend value as well (to prevent cache mismatch)
        region.__dict__.pop('_actual_backend', None)
        region.configure_from_config(settings, prefix='')

    # Since get_region() returns an unconfigured region for *any* name you
    # pass in, we make sure that all used regions are configured now.
    for region in CACHE_REGIONS.values():
        if 'backend' not in region.__dict__:
            raise dogpile.cache.exception.RegionNotConfigured(
                'Region %r used in python code, but not configured' % region.name
            )


def _parse_dogpile_cache_settings(settings):
    # XXX Woefully incomplete. This only supports pylibmc, and our specific
    # use-case: all regions use the same memcache settings.
    if settings.get('dogpile_cache.pylibmc_url'):
        settings['dogpile_cache.arguments.url'] = settings['dogpile_cache.pylibmc_url'].split(';')
        del settings['dogpile_cache.pylibmc_url']

    for key in [
        'dogpile_cache.arguments.lock_timeout',
        'dogpile_cache.arguments.memcache_expire_time',
    ]:
        if settings.get(key):
            settings[key] = int(settings[key])

    behaviors = {}
    behavior_prefix = 'dogpile_cache.pylibmc_behavior.'
    to_remove = []
    for key, value in settings.items():
        if not key.startswith(behavior_prefix):
            continue
        behaviors[key.replace(behavior_prefix, '')] = value
        to_remove.append(key)
    if behaviors:
        from pyramid_dogpile_cache2 import pylibmc

        settings['dogpile_cache.arguments.behaviors'] = pylibmc.convert(behaviors)
        for key in to_remove:
            del settings[key]



"""
Adapted from https://github.com/moriyoshi/pyramid_dogpile_cache/blob/0.0.4/pyramid_dogpile_cache/__init__.py

Copyright (c) 2014 Moriyoshi Koizumi

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.
"""


def build_dogpile_region_settings(settings):
    region_settings_dict = {}
    prefix = 'dogpile_cache.'

    # pass 1: retrieve settings and put them into the dictionary
    for key, value in settings.items():
        if key.startswith(prefix):
            region_name, dot, param_name = key[len(prefix) :].partition('.')
            if not dot:
                if region_name == 'regions':
                    # dogpile_cache.regions = foo, bar ...
                    for region_name in value.split(','):
                        region_name = region_name.strip()
                        if region_name in ('regions', 'arguments'):
                            raise ValueError('region name %s is not allowed' % region_name)
                        if region_name and region_name not in region_settings_dict:
                            region_settings_dict[region_name] = {}
                    continue
                else:
                    # dogpile_cache.backend = ...
                    # dogpile_cache.expiration_time = ...
                    param_name = region_name
                    region_name = ''
                    if param_name == 'name':
                        raise ValueError(
                            'parameter %s is not allowed for the default cache settings' % key
                        )
            elif region_name == 'arguments':
                # dogpile_cache.arguments.host = localhost ...
                param_name = region_name + '.' + param_name
                region_name = ''

            region_settings = region_settings_dict.get(region_name)
            if not region_settings:
                region_settings = region_settings_dict[region_name] = {}
            region_settings[param_name] = value

    # pass 2: combine region-specific settings with defaults
    default_settings = region_settings_dict.get('', {})
    for region_name, region_settings in region_settings_dict.items():
        if region_name == '':
            # skip the default
            continue

        # merge the default into region-specific settings
        for key, value in default_settings.items():
            if key not in region_settings:
                region_settings[key] = value

    region_settings_dict.pop('', None)

    return default_settings, region_settings_dict
